import re
from calendar import monthrange
from datetime import datetime, date, timedelta, timezone
from typing import Union, Optional, Generator

DATE_FORMAT = "%Y-%m-%d"
TIME_FORMAT = "%H:%M:%S"
DATETIME_FORMAT = f"{DATE_FORMAT} {TIME_FORMAT}"

# ------------------------ Taken from django dateparse ----------------------- #

datetime_re = re.compile(
    r'(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})'
    r'[T ](?P<hour>\d{1,2}):(?P<minute>\d{1,2})'
    r'(?::(?P<second>\d{1,2})(?:[\.,](?P<microsecond>\d{1,6})\d{0,6})?)?'
    r'\s*(?P<tzinfo>Z|[+-]\d{2}(?::?\d{2})?)?$'
)

date_re = re.compile(
    r'(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})$'
)


def get_fixed_timezone(offset):  # pragma no cover
    """Return a tzinfo instance with a fixed offset from UTC."""
    if isinstance(offset, timedelta):
        offset = offset.total_seconds() // 60
    sign = '-' if offset < 0 else '+'
    hhmm = '%02d%02d' % divmod(abs(offset), 60)
    name = sign + hhmm
    return timezone(timedelta(minutes=offset), name)


def parse_datetime(value):  # pragma no cover
    """Parse a string and return a datetime.datetime.

    This function supports time zone offsets. When the input contains one,
    the output uses a timezone with a fixed offset from UTC.

    Raise ValueError if the input is well formatted but not a valid datetime.
    Return None if the input isn't well formatted.
    """
    try:
        return datetime.fromisoformat(value)
    except ValueError:
        if match := datetime_re.match(value):
            kw = match.groupdict()
            kw['microsecond'] = kw['microsecond'] and kw['microsecond'].ljust(6, '0')
            tzinfo = kw.pop('tzinfo')
            if tzinfo == 'Z':
                tzinfo = timezone.utc
            elif tzinfo is not None:
                offset_mins = int(tzinfo[-2:]) if len(tzinfo) > 3 else 0
                offset = 60 * int(tzinfo[1:3]) + offset_mins
                if tzinfo[0] == '-':
                    offset = -offset
                tzinfo = get_fixed_timezone(offset)
            kw = {k: int(v) for k, v in kw.items() if v is not None}
            return datetime(**kw, tzinfo=tzinfo)


def parse_date(value):  # pragma no cover
    """Parse a string and return a datetime.date.

    Raise ValueError if the input is well formatted but not a valid date.
    Return None if the input isn't well formatted.
    """
    try:
        return date.fromisoformat(value)
    except ValueError:
        if match := date_re.match(value):
            kw = {k: int(v) for k, v in match.groupdict().items()}
            return date(**kw)


# ---------------------------------------------------------------------------- #

def date_range(start_date: date, end_date: date) -> Generator[date, None, None]:
    """
    Generator for yield-ing date objects in a given range. If start_date == end_date
    it still yields just one time the value. INCLUDES end_date!
    Args:
        start_date: date obj from where to start the generator e.g: date(2021, 1, 1)
        end_date: date obj from where to start the generator e.g: date(2021, 1, 1),
                    included in the end
    Returns:
        Generator for date range
    Examples:
        >>> a = date_range(date(2021, 1, 1), date(2021, 2, 1))
        >>> print(type(a))
        <class 'generator'>
        >>> sum(1 for _ in a)
        32
    """
    for n in range(int((end_date - start_date).days + 1)):
        yield start_date + timedelta(n)


def last_day_of_month(month: int, year: int) -> date:
    """
    Take the month and year and return a date representing
    the last day of the given month. Supports leap years.
    Args:
        year: Int representing the year
        month: Int between 1-12 representing the month
    Returns:
        datetime obj representing the last day of month
    Examples:
        >>> last_day_of_month(1, 2022)
        datetime.date(2022, 1, 31)
        >>> last_day_of_month(2, 2021)
        datetime.date(2021, 2, 28)
        >>> last_day_of_month(2, 2020)
        datetime.date(2020, 2, 29)
    """
    current_month_last_day = monthrange(year, month)[1]
    return date(year, month, current_month_last_day)


def get_previous_end_of_month(reference_date: date) -> date:
    """
    Take a date object and return the last date of the previous month
    Eg: date(2021, 2, 6) -> returns date(2021, 1, 31)
    Args:
        reference_date: date object
    Returns:
         date object
    Examples:
        >>> get_previous_end_of_month(date(2021, 2, 6))
        datetime.date(2021, 1, 31)
        >>> get_previous_end_of_month(date(2021, 1, 1))
        datetime.date(2020, 12, 31)
    """
    return reference_date.replace(day=1) - timedelta(days=1)


def date_to_datetime(date_to_convert: date) -> datetime:
    """
    Convert from date obj to datetime obj
    Args:
        date_to_convert: date obj to convert
    Returns:
        datetime obje
    Examples:
          >>> date_to_datetime(date(2021, 5, 23))
          datetime.datetime(2021, 5, 23, 0, 0)
    """
    return datetime(
        year=date_to_convert.year,
        month=date_to_convert.month,
        day=date_to_convert.day
    )


def get_datetime(
        value: Optional[Union[date, datetime, str]],
        raise_error=False
) -> Optional[datetime]:
    """
    Convert a given value to a datetime.
    Args:
        raise_error: flag to raise error if return is None or not
        value: to be converted. Can be date/datetime obj as well as str formated in date/datetime
    Returns:
        datetime obj
    Raises:
        ValueError: If raise_error flag is True and parsed_datetime is None
    Examples:
        >>> get_datetime(date(2021, 1, 1))
        datetime.datetime(2021, 1, 1, 0, 0)
        >>> get_datetime(datetime(2021, 1, 1, 0, 2))
        datetime.datetime(2021, 1, 1, 0, 2)
        >>> get_datetime('2021-1-1')
        datetime.datetime(2021, 1, 1, 0, 0)
        >>> get_datetime('2021-20-20-20-20') is None
        True
        >>> get_datetime(None) is None
        True
        >>> get_datetime('2021-20-20-20-20', raise_error=True)
        Traceback (most recent call last):
        ...
        ValueError: Invalid datetime 2021-20-20-20-20
    """
    if isinstance(value, datetime):
        return value
    if isinstance(value, date):
        return date_to_datetime(value)

    # A common date is in the form "2020-01-01", 10 characters
    if value is not None:
        if len(value) > 10:
            parsed_datetime = parse_datetime(value)
        else:
            parsed_date = parse_date(value)
            parsed_datetime = date_to_datetime(parsed_date) if parsed_date else None
    else:
        parsed_datetime = None

    if parsed_datetime is None and raise_error:
        raise ValueError(f"Invalid datetime {value}")
    return parsed_datetime


def get_date(
        value: Optional[Union[date, datetime, str]],
        raise_error=False
) -> Optional[date]:
    """
    Convert a given value to a date.
    Args:
        raise_error: flag to raise error if return is None or not
        value: to be converted. Can be date/datetime obj as well as str formatted in date/datetime
    Returns:
        date obj
    Raises:
        ValueError: If raise_error flag is True and parsed_date is None
    Examples:
        >>> get_date(date(2021, 1, 1))
        datetime.date(2021, 1, 1)
        >>> get_date(datetime(2021, 1, 1, 0, 2))
        datetime.date(2021, 1, 1)
        >>> get_date('2020-01-01 13:12:13')
        datetime.date(2020, 1, 1)
        >>> get_date('sadasadasdas') is None
        True
        >>> get_date(None) is None
        True
        >>> get_date('2021-20-20-20-20', raise_error=True)
        Traceback (most recent call last):
        ...
        ValueError: Invalid date 2021-20-20-20-20
    """
    if isinstance(value, datetime):
        return value.date()
    if isinstance(value, date):
        return value

    if value is not None:
        # A common date is in the form "2020-01-01", 10 characters
        if len(value) > 10:
            parsed_date = parse_datetime(value)
            parsed_date = parsed_date.date() if parsed_date else None
        else:
            parsed_date = parse_date(value)
    else:
        parsed_date = None

    if parsed_date is None and raise_error:
        raise ValueError(f"Invalid date {value}")

    return parsed_date


def difference_months(start_date: date, end_date: date) -> int:
    """
    Get the difference in months between two date objects.
    Args:
        start_date (date): the beginning of the period
        end_date (date): the end of the period
    Returns:
        difference as int
    Examples:
        >>> difference_months(date(2020, 10, 1), date(2021, 10, 1))
        12
        >>> difference_months(date(2021, 10, 1), date(2021, 10, 1))
        0
        >>> difference_months(date(2021, 10, 1), date(2021, 8, 1))
        -2
    """
    return (end_date.year - start_date.year) * 12 + (end_date.month - start_date.month)


def get_last_date_of_week(reference_date: Optional[date] = None) -> date:
    """
    1.Give a reference date or get the current date as a start point.
    2.Return the last day of the week the given date is in.
    Args:
        reference_date: day in the week to start the search for
    Returns:
        last day of the week
    Examples:
        >>> get_last_date_of_week(date(2022, 1, 27))
        datetime.date(2022, 1, 30)
    """
    if not reference_date:  # pragma: no cover
        reference_date = date.today()
    return reference_date + timedelta(days=6 - reference_date.weekday())



import hashlib
import re
from decimal import Decimal
from typing import Optional, List, Any

from python_utils.types_hinting import NumberIFD


def format_percent(value: Optional[NumberIFD]):
    """
    Get value and return a % repr

    Args:
        value: value to be used for % repr
    Returns:
        A % repr of the passed value
    Examples:
        >>> format_percent(13)
        '1300%'
        >>> format_percent(0.13)
        '13.0%'
        >>> format_percent(Decimal('0.123456'))
        '12.35%'
    """
    return f"{round(value * 100, 2)}%"


def format_currency(value: Optional[NumberIFD], symbol: str) -> Optional[str]:
    """
    Get value and number and return a currency representation

    Args:
        value:
        symbol: string representing the symbol of currency
    Returns:
        string representing currency and value
    Examples:
        >>> format_currency(120.1234, 'EUR')
        'EUR120.12'
        >>> format_currency(Decimal('120.1234'), 'EUR')
        'EUR120.12'
        >>> format_currency(None, 'EUR') is None
        True
    """
    if not value:
        return None
    return f"{symbol}{round(value, 2):,}"


def human_string(text: str) -> str:
    """
    Transform text to human string

    Args:
        text: string to be converted
    Returns:
        converted string
    Examples:
        >>> human_string('test_str')
        'Test Str'
        >>> human_string("")
        ''
    """
    if not text:
        return ""
    return ' '.join(word.title() for word in text.split('_'))


camel_case_pattern = re.compile(r"(?<!^)(?=[A-Z])")


def camel_case_to_snake_case(text: str) -> str:
    """
    Convert camel case string to snake case, Substitute spaces with _

    Args:
        text: string to be converted
    Returns:
        converted string
    Examples:
        >>> camel_case_to_snake_case('TestStrHello')
        'test_str_hello'
        >>> camel_case_to_snake_case('Test Str Hello  dummy CAPITAL12')
        'test__str__hello__dummy__c_a_p_i_t_a_l12'
    """
    return camel_case_pattern.sub("_", text.replace(" ", "_")).lower()


# Indicates 1 or more spaces
spaces_pattern = re.compile(r" +")


def human_string_to_snake_case(text: str) -> str:
    """
    Transform text from human string to snake case

    Args:
        text: string to be converted
    Returns:
        converted string
    Examples:
        >>> human_string_to_snake_case('Test Str TEST')
        'test_str_test'
    """
    return spaces_pattern.sub("_", text.strip().lower())


def extract_with_pattern(text: str, pattern: str) -> Optional[str]:
    """
    Return the first found result of a given pattern in the string or None if not found
    E.g: ('Test _t Test _t', '_t') -> '_t'

    Args:
        text: string to find the match
        pattern: pattern against which the string is trying to match
    Returns:
        First result of given pattern or None
    Examples:
        >>> extract_with_pattern('Test _t Test _t', '_t')
        '_t'
        >>> extract_with_pattern('Test _t Test _t', 'A') is None
        True
        >>> extract_with_pattern('Test 123', '\\d+')
        '123'
    """
    compiled_pattern = re.compile(pattern)
    results = compiled_pattern.search(text)
    if results:
        return results.group(0)


def create_hash(data: List[Any]) -> str:
    """
    Given a list of params calculate a hash based on their values

    Args:
        data: list of params to be used for the hash
    Returns:
        calculated hash value
    Examples:
        >>> create_hash(['test', 'hash', 1])
        '5ca1b365312e58a36bb985fc770a490b'
    """
    merged_data = '-'.join([str(rec) for rec in data])
    return hashlib.md5(merged_data.encode('utf-8')).hexdigest()



from sqlite3 import Cursor
from typing import List, Dict


def fetch_all(cursor: Cursor) -> List[Dict]:
    """
    Returns all rows from a cursor as a list of dictionaries.

    Args:
        cursor: Cursor object storing the data to be fetched
    Returns:
         list with all the records retrieved from the db formatted as dict
    """
    desc = cursor.description
    return [dict(zip([col[0] for col in desc], row)) for row in cursor.fetchall()]



import logging
import sys
import time
from typing import Dict, Union, Optional, IO, TypeVar
from typing import List, Any
from typing import Tuple

from django.core.exceptions import ObjectDoesNotExist
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models import Model
from django.db.models import Sum, Min, Avg, Max, Subquery, Case, When, ExpressionWrapper, F
from django.db.models.fields.files import FieldFile
from django.forms import model_to_dict
from django.utils import timezone

from python_utils.imports import import_optional_dependency
from python_utils.db import fetch_all

Model_T = TypeVar('Model_T', bound=Model)


class ExtendedEncoder(DjangoJSONEncoder):  # pragma no cover
    """Custom JSON Encoder that processes fields like date etc."""

    def default(self, o):
        if isinstance(o, Model):
            return model_to_dict(o)
        elif isinstance(o, FieldFile):
            return o.name
        return super().default(o)


def record_to_dict(record: Model_T, exclude: List = None) -> Dict:
    """
    Transform a django record to a dict.

    Args:
        record: django object to be transformed to dict
        exclude: list of keys/fields to exclude in the final dict
    Returns:
        dict with model's field:value pairs
    """

    if exclude is None:
        exclude = []

    initial_data = {
        key: value
        for key, value in record.__dict__.items() if not (key.startswith('_') or key in exclude)
    }

    # Replace choices with their human representation
    for field, value in initial_data.items():
        if value is None:
            continue

        if get_field_display := getattr(record, f'get_{field}_display', None):
            initial_data[field] = get_field_display()

    return initial_data


def get_choices_list_and_dict(choices) -> Tuple[List, Dict]:
    """
    From a Choices object return a list and dict of these choices.
    Args:
        choices: Choices object to get the representations from
    Returns:
        tuple of list with human_value and dict-> {human_value: value}
    Examples:
        >>> from model_utils import Choices
        >>> get_choices_list_and_dict(Choices((1, 'hello', 'HELLO')))
        ([('HELLO', 'HELLO')], {'HELLO': 1})
    """
    import_optional_dependency('model_utils', dependency='django-model-utils')

    choices_list = [(human_value, human_value) for _, human_value in choices]
    choices_dict = {human_value: value for value, human_value in choices}
    return choices_list, choices_dict


def perform_query(
        sql_query: str,
        params: Optional[List] = None,
        expected_results=True,
) -> Optional[List[Dict]]:
    """
    Execute a sql statement to the database and get the results as a list of dict or None
    if no result is expected

    Args:
        sql_query: string representing the query to be executed
        params: list with params to be passed to the query
        expected_results: indicating whether the query executed will return or not results
    Returns:
        List of dict with the fetched records or None if expected_result=False
    """
    from django.db import connection

    with connection.cursor() as cursor:
        if params:
            cursor.execute(sql_query, tuple(params))
        else:
            cursor.execute(sql_query)
        if expected_results:
            return fetch_all(cursor)


def get_total(records: models.QuerySet, aggregation_field: str) -> Optional[int]:
    """
    Calculate the total value of the aggregation field for the provided records

    Args:
        records: queryset with the records to get the total sum of the passed field
        aggregation_field: string representing the field we want to get the sum of
    Returns:
        Total sum of the field or None if all values of field are empty or empty queryset
    """
    return records.aggregate(total=Sum(aggregation_field)).get("total")


def get_min(records: models.QuerySet, aggregation_field: str) -> Optional[int]:
    """
    Calculate the min value of the aggregation field for the provided records

    Args:
        records: queryset with the records to get the min value of the passed field
        aggregation_field: string representing the field we want to get the min value of
    Returns:
        Min value found of the field or None if all values of field are empty or empty queryset
    """
    return records.aggregate(min=Min(aggregation_field)).get("min")


def get_average(records: models.QuerySet, aggregation_field: str) -> Optional[int]:
    """
    Calculate the average value of the aggregation field for the provided records
    Args:
        records: queryset with the records to get the average of the passed field
        aggregation_field: string representing the field we want to get the average of
    Returns:
        Average of the field or None if all values of field are empty or empty queryset
    """
    return records.aggregate(avg=Avg(aggregation_field)).get("avg")


def get_max(records: models.QuerySet, aggregation_field: str) -> Optional[int]:
    """
    Calculate the max value of the aggregation field for the provided records

    Args:
        records: queryset with the records to get the max of the passed field
        aggregation_field: string representing the field we want to get the max of
    Returns:
        Max of the field or None if all values of field are empty or empty queryset
    """
    return records.aggregate(max=Max(aggregation_field)).get('max')


def compute_weighted_average(
        records: models.QuerySet,
        field: str,
        weight: str = "amount",
        is_field_subquery=False,
        is_weight_subquery=False,
        output=models.FloatField,
) -> models.QuerySet:
    field = Subquery(field) if is_field_subquery else F(field)
    weight = Subquery(weight) if is_weight_subquery else F(weight)
    return records.aggregate(
        weight_sum=Sum(weight),
        average=Case(
            When(weight_sum=0, then=0),
            default=ExpressionWrapper(
                Sum(field * weight) / Sum(weight), output_field=output()
            ),
            output_field=output(),
        ),
    ).get("average")


def compute_grouped_weighted_average(
        records: models.QuerySet,
        group_by: List[str],
        field: str,
        weight: str = "amount",
        label: str = "avg",
        output=models.FloatField
) -> models.QuerySet:
    annotation = {
        label: ExpressionWrapper(
            Sum(F(field) * F(weight)) / Sum(F(weight)), output_field=output()
        )
    }
    return records.values(*group_by).annotate(**annotation)


def compute_grouped_weighted_average_subquery(
        records: models.QuerySet,
        group_by: List[str],
        field_subquery: Subquery,
        weight_subquery: Subquery,
        label: str = "avg",
        output=models.FloatField,
) -> models.QuerySet:
    annotation = {
        label: ExpressionWrapper(
            Sum(field_subquery * weight_subquery) / Sum(weight_subquery),
            output_field=output()
        )
    }
    return records.values(*group_by).annotate(**annotation)


def update_record(record: Model_T, save=True, **data) -> Model_T:
    """
    Update a record with given attributes and return it.
    If save=True also commit to the database the record with the changes

    Args:
        record: db record to be updated
        save: flag whether to save the record
        **data: data to be changed mapped with the record fields as kwargs
    Returns:
        New updated record
    """
    if data:
        for key, value in data.items():
            setattr(record, key, value)
        if save:
            record.save()
    return record


def get_choice_value(choices, human_string: str) -> Optional[Union[int, str]]:
    """
    Get database representation of a choice for a human-readable value

    Args:
        choices: model_utils Choice object to get the db value from
        human_string: string representing the human-readable value
    Returns:
        db representation of choice field
    """

    import_optional_dependency('model_utils', dependency='django-model-utils')

    for value, representation in choices:
        if representation == human_string:
            return value
    return None


def get_or_none(records: models.QuerySet, *args, **kwargs) -> Optional[Model_T]:
    """
    Wrapper around queryset.get to return None if no record is found

    Args:
        records: queryset of objects to get() from
        *args: args to be passed to get() method
        **kwargs: kwargs to be passed to get() method
    Returns:
        record object if found or None
    """
    try:
        return records.get(*args, **kwargs)
    except ObjectDoesNotExist:
        return None


def ids(queryset: models.QuerySet, attr='id') -> List[Any]:
    """
    Given a queryset and an attr, get all distinct values of the attr as a list

    Args:
        queryset: queryset with the objects we want to get the data from
        attr: string representing the field of the model we want to get the distinct values
    Returns:
        List of distinct values
    Examples:
        ids(Invoice.objects.filter(id__gt=50), 'id') \n
        [51, 52, 53, 54, 55 ...]
    """
    return list(queryset.order_by(attr).values_list(attr, flat=True).distinct())


def get_field_config(field: models.Field) -> Dict:
    """
    Get the configuration of a Django field, in the form: {'name': ..., 'type': ..., 'choices': ..., 'description: ...}
    Available types are ["string", "integer", "float", "amount", "datetime", "date", "selection"]
    selection type is used when the value of the field supports choices or when it is a foreign key.
    When the type is selection, the config of the field is in the form
    {
        "type": "selection",
        "choices": [{"id": 1, "name": "Choice"}]
    }

    Args:
        field: The field for which the config is required
    Returns:
        The config of the field.
    """

    field_type = None
    choices = None

    if getattr(field, "choices", None):
        field_type = "selection"
        choices = [{"id": value, "name": label} for value, label in field.choices]
    elif isinstance(field, models.ForeignKey):
        field_type = "selection"
        choices = [
            {"id": record.id, "name": record.name}
            for record in field.related_model.objects.all()
        ]

    elif isinstance(field, (models.CharField, models.TextField)):
        field_type = "string"
    elif isinstance(field, models.IntegerField):
        field_type = "integer"
    elif isinstance(field, models.FloatField):
        field_type = "float"
    elif isinstance(field, models.DecimalField):
        field_type = "amount"
    elif isinstance(field, models.DateTimeField):
        field_type = "datetime"
    elif isinstance(field, models.DateField):
        field_type = "date"
    elif isinstance(field, models.JSONField):
        field_type = "json"
    elif isinstance(field, models.BooleanField):
        field_type = "bool"

    field_config = {
        "name": field.name,
        "type": field_type,
        "description": field.help_text if hasattr(field, "help_text") else "",
    }
    if choices:
        field_config["choices"] = choices

    return field_config


def get_model_fields_config(
        model: models.Model.__class__, excluded_fields: List[str] = None, read_only_fields: List[str] = None
) -> List[Dict]:
    """
    Get the configurations for the fields of a model.

    Args:
        model: The model for which the fields config is required
        excluded_fields: A list of field names which should be excluded
        read_only_fields: A list of field names which should be considered as read_only
    Returns:
        A List of dictionaries which represent the field configurations
    """

    if not excluded_fields:
        excluded_fields = []
    if not read_only_fields:
        read_only_fields = []

    fields_config = []

    # Process all fields that can contain values
    for field in [
        field for field in model._meta.get_fields() if field.name not in excluded_fields
    ]:
        field_config = get_field_config(field)
        field_config["input"] = field.name not in read_only_fields
        fields_config.append(field_config)

    return fields_config


def get_fields_config_and_values(
        record: models.Model,
        excluded_fields: List[str] = None,
        read_only_fields: List[str] = None
) -> List[Dict]:
    fields_config = get_model_fields_config(
        type(record), excluded_fields, read_only_fields
    )

    # Add value for each field
    for field_config in fields_config:
        value = getattr(record, field_config["name"])
        if isinstance(value, models.Model):
            value = value.id
        field_config["value"] = value

    return fields_config


def get_model_field_names(model: Union[models.Model.__class__, Model_T]) -> List[str]:
    """
    Iterate over a model fields and return a list of them

    Args:
        model: model object to iterate over and get fields
    Returns:
        list[str] of all the fields
    Examples:
        Invoice:
            code=models.CharField(max_length=100) \n
            amount=models.Integer() \n

        get_model_field_names(Invoice) \n
        [id, code, amount]
    """
    return [field.name for field in model._meta.get_fields()]


def safe_bulk_create(
        model: models.Model.__class__,
        records: List[Model_T],
        refresh_fields: List[str] = None,
        batch_size: int = 500
):
    """
    Given a model and a list of records, bulk_create them and also set the
    refresh fields (foreign keys) if any.

    Args:
        model: The Model to which the records belong
        records: list of objects to be commited to the database
        refresh_fields: list with foreign keys to be updated
        batch_size: how many records we want to insert with a single query
    """
    if not records:
        return

    if refresh_fields:
        for record in records:
            for field in refresh_fields:
                try:
                    setattr(record, f"{field}_id", getattr(record, field).id)
                except AttributeError:
                    continue

    model.objects.bulk_create(records, batch_size=batch_size)


def safe_bulk_update(
        model: models.Model.__class__,
        records: List[Model_T],
        fields: List[str],
        batch_size: int = 500,
        refresh_fields: List[str] = None,
        updated_at_field: str = 'updated_at'
):
    """
    Given a model and a list of records, bulk_update them and also set the
    refresh fields (foreign keys) if any. Supports auto now fields to be updated as well

    Args:
        model: Model.__class__ obj
        records: list of records to be created
        fields: list of fields needed to be updated
        refresh_fields: list with foreign keys to be updated
        batch_size: how many records we want to insert with a single query
        updated_at_field: field datetime/date repr updated at moment
    """
    if not records:
        return

    if refresh_fields:
        for record in records:
            for field in refresh_fields:
                try:
                    setattr(record, f"{field}_id", getattr(record, field).id)
                except AttributeError:
                    continue

    # Auto now fields are not updated automatically in bulk operations.
    if hasattr(model, updated_at_field) and (updated_at_field not in fields):
        # This kind of update supports both `set` and `list`
        fields = [*fields, updated_at_field]
        timestamp = timezone.now()
        for record in records:
            setattr(record, updated_at_field, timestamp)
    model.objects.bulk_update(records, fields, batch_size=batch_size)


logger = logging.getLogger(__name__)


def call_procedure(procedure_name: str, params: List[Any] = None):
    """
    Execute an SQL procedure in the db.

    Args:
        procedure_name: the name of the procedure to be executed
        params: list of the parameters to pass to the procedure
    """
    temp_time = time.time()
    logger.info(f"Started Procedure {procedure_name}")
    query = f"call {procedure_name}"
    if params:
        query += f"({(',%s' * len(params))[1:]});"
    else:
        query += "();"
    perform_query(
        sql_query=query,
        params=params,
        expected_results=False
    )
    logger.info(f"Finished Procedure {procedure_name} in {time.time() - temp_time}")


def get_id_field_map(queryset: models.QuerySet, field: str) -> Dict[int, Any]:
    """
    Given a queryset get a mapping of the form \n
    {
        <id>: <field>, \n
        ...
    }

    Args:
        queryset:   Django models.QuerySet Instance
        field:      Field that needs to be mapped, for ex. `name`, `identifier`, etc.
    Returns:
        Mapping with ids as keys and the given field as values.
    """
    return {
        record[0]: record[1]
        for record in queryset.values_list('id', field)
    }


def generate_file_from_buffer(file_buffer_io: IO[str], content_type: str, filename="file", encoding=None):
    field_type = 'FileField'
    size = sys.getsizeof(file_buffer_io)
    return InMemoryUploadedFile(file_buffer_io, field_type, filename, content_type, size, encoding)



from importlib import import_module


def import_optional_dependency(module_to_import: str, dependency: str = None):
    """
    Attempt to import a module from an external dependency.
    Raise an error if that dependency is not installed.
    Args:
        module_to_import: A string containing the module to be imported
        dependency: Information about the dependency required
    Raises:
        ModuleNotFoundError: If the dependency is not installed
    Examples:
        >>> math = import_optional_dependency('math')
        >>> pyyaml = import_optional_dependency("pyyaml")
        Traceback (most recent call last):
        ...
        ModuleNotFoundError: pyyaml is required to use this function.
    """
    try:
        return import_module(module_to_import)
    except ModuleNotFoundError:
        raise ModuleNotFoundError(f"{dependency or module_to_import} is required to use this function.")



from datetime import date, datetime
from typing import List, Dict, Union, Literal

import pandas as pd

from python_utils.time import get_date, last_day_of_month


def unique_not_null_values(df: pd.DataFrame, column: str) -> List:
    """
    Get a list with unique values from the values of a dataframe column
    1. Filter the df to keep rows with notnull values in the given column
    2. Return a list with all the values of the column appearing only once

    Args:
        df: dataframe object we are getting the unique values from
        column: the column name on which we are looking for the unique values
    Returns:
        list with all unique values
    Examples:
        >>> test_df = pd.DataFrame({'col1': [1, None, 1, 3, 2], 'col2': [3, 7, 4, 1, 3]})
        >>> unique_not_null_values(test_df, 'col1')
        [1.0, 3.0, 2.0]
    """
    return df[df[column].notnull()][column].unique().tolist()


def rename_and_replace_column_information(df: pd.DataFrame, data: Dict) -> pd.DataFrame:
    """
    We use this mostly when we serialize the data from postgres.
    It replaces the column values with a dictionary and also renames the column names.
    Performs a rename of the columns

    Args:
        df: Pandas Data Frame
        data: data that maps columns and replacement/mapping as in example:
            {
            'column_in_pandas': {
                'replace_to': 'new_column_name',
                'map': {
                    1: 2,
                    2: 3
                }
            }
    Returns:
        New Modified Pandas Dataframe.
    Examples:
        >>> test_df = pd.DataFrame({'col1': ['a', 'b'], 'col2': [1, 2]})
        >>> rename_and_replace_column_information(test_df, {'col1': {'replace_to': 'new', 'map': {'a': 1}}})
          new  col2
        0   1     1
        1   b     2
        >>> test_df = pd.DataFrame({'col1': ['a', 'b'], 'col2': [1, 2]})
        >>> rename_and_replace_column_information(test_df, {})
          col1  col2
        0    a     1
        1    b     2
    """
    map_data = {key: value.get('map') for key, value in data.items()}
    if map_data:
        df.replace(map_data, inplace=True)

    rename_columns = {key: value.get('replace_to') for key, value in data.items()}
    if rename_columns:
        df.rename(columns=rename_columns, inplace=True)
    return df


def get_dates_between(
        start_date: Union[date, datetime, str],
        end_date: Union[date, datetime, str],
        inclusive: Literal["both", "neither", "left", "right"] = "both"
) -> list[date]:
    """
    1. Give start and end dates.
    2. Set inclusive if you want, by default the value is both
    3. Return list of dates
    Args:
        start_date (date, datetime, str): The left bound for the date generation by pandas
        end_date (date, datetime, str): The right bound for the date generation by pandas
        inclusive: Include boundaries; Whether to set each bound as closed or open
    Returns:
        list of date obj representing the dates in between the start and end
    Raises:
        ValueError: if one of passed dates is str and not well formatted to be converted to date obj
        Examples:
        >>> get_dates_between(date(2021, 2, 1), date(2021, 2, 3))
        [datetime.date(2021, 2, 1), datetime.date(2021, 2, 2), datetime.date(2021, 2, 3)]
        >>> get_dates_between(datetime(2021, 2, 1, 0, 0), datetime(2021, 2, 3, 0, 0))
        [datetime.date(2021, 2, 1), datetime.date(2021, 2, 2), datetime.date(2021, 2, 3)]
        >>> get_dates_between("2021-2-1", "2021-2-3")
        [datetime.date(2021, 2, 1), datetime.date(2021, 2, 2), datetime.date(2021, 2, 3)]
        >>> get_dates_between(date(2021, 2, 1), date(2021, 2, 3), inclusive="right")
        [datetime.date(2021, 2, 2), datetime.date(2021, 2, 3)]
        >>> get_dates_between("2021-2-1-1-1", "2021-3-1-1-1-1")
        Traceback (most recent call last):
        ...
        ValueError: Invalid date 2021-2-1-1-1
    """
    start_date = get_date(start_date, raise_error=True)
    end_date = get_date(end_date, raise_error=True)
    dates_range = pd.date_range(start=start_date, end=end_date, inclusive=inclusive).tolist()
    return [date(d.year, d.month, d.day) for d in dates_range]


def get_months_between_dates(
        start_date: Union[date, datetime, str],
        end_date: Union[date, datetime, str],
        inclusive: Literal["both", "neither", "left", "right"] = "both"
) -> List[date]:
    """
    Get a timestamp for each month between the two dates.
    The start_date is appended because the pandas list begins from the next month

    Args:
        start_date (date, datetime, str): The left bound for the date generation by pandas
        end_date (date, datetime, str): The right bound for the date generation by pandas
        inclusive: Include boundaries; Whether to set each bound as closed or open
    Returns:
        list of date obj representing the end date of each month in between
    Raises:
        ValueError: if one of passed dates is str and not well formatted to be converted to date obj
    Examples:
        >>> get_months_between_dates(date(2021, 2, 1), date(2021, 3, 1))
        [datetime.date(2021, 2, 28), datetime.date(2021, 3, 31)]
        >>> get_months_between_dates(datetime(2021, 2, 1, 0, 0), datetime(2021, 3, 1, 1, 1))
        [datetime.date(2021, 2, 28), datetime.date(2021, 3, 31)]
        >>> get_months_between_dates("2021-2-1", "2021-3-1")
        [datetime.date(2021, 2, 28), datetime.date(2021, 3, 31)]
        >>> get_months_between_dates(date(2021, 2, 1), date(2021, 3, 1), inclusive="right")
        [datetime.date(2021, 3, 31)]
        >>> get_months_between_dates("2021-2-1-1-1", "2021-3-1-1-1-1")
        Traceback (most recent call last):
        ...
        ValueError: Invalid date 2021-2-1-1-1
    """
    start_date = get_date(start_date, raise_error=True)
    end_date = get_date(end_date, raise_error=True)
    dates_range = pd.date_range(start_date, end_date, freq="MS", inclusive=inclusive).to_list()
    return [
        last_day_of_month(year=cur_month.year, month=cur_month.month)
        for cur_month in dates_range
    ]



import operator
from datetime import datetime, timezone
from decimal import Decimal
from typing import (
    Callable, Hashable, Generator, Iterable, List, Any, Dict, Optional, Union
)

from python_utils.math import are_equal
from python_utils.types_hinting import DictOrObject


def safe_find(records: Iterable, filter_func: Callable) -> Optional[Any]:
    """
    Wrapper around the filter function, to return only the first record when
    there is a match, or None instead.

    Args:
        records: List to iterate over
        filter_func: Function that will be appliead on each record
    Returns:
        First occurrence found or None
    Examples:
        >>> safe_find([{'a': 'T', 'b': 'H'}, {'a': 'F'}], lambda rec: rec.get('a') == 'F')
        {'a': 'F'}
        >>> safe_find([1, 2, 3], lambda rec: rec > 0)
        1
        >>> safe_find([{'a': 'T'}], lambda rec: rec.get('a') == 'F') is None
        True
    """
    try:
        return next(filter(filter_func, records))
    except StopIteration:
        return None


def find_by(records: List[Dict], attr: Hashable, value: Any) -> Optional[Dict]:
    """
    Given a list of dicts and an attribute, value, return the first occurrence
    where dict[attr] == value or None if no occurrence accomplishes the condition

    Args:
        records: list of dicts to iterate over
        attr: the attr we're trying to check for
        value: value to make the check against
    Returns:
        Dict record if found or None
    Examples:
        >>> find_by([{'a': 'T', 'b': 'H'}, {'a': 'F'}], attr='b', value='H')
        {'a': 'T', 'b': 'H'}
        >>> find_by([{'a': 'T', 'b': 'H'}, {'a': 'F'}], attr='a', value='Hello') is None
        True
    """
    return safe_find(records, lambda rec: value == rec.get(attr))


def filter_dicts(records: List[Dict], as_list=True, **filters) -> Union[List[Dict], Iterable[Dict]]:
    """
    Pass a list of dicts and filters as kwargs to get all filtered records as list or generator

    Args:
        records: list of dicts to iterate and filter from
        as_list: Flag that shows if found records should be returned as list
        **filters: kwargs to be used for filtering as key:value pairs
    Returns:
        Filtered records as filter obj or list
    Examples:
        >>> test_d = [{'a': 1, 'b': 2}, {'a': 0, 'b': 3}, {'a': 1, 'd': 3}, {'a': 2, 'd': 3}]
        >>> filter_dicts(test_d,  a=1, d=3)
        [{'a': 1, 'd': 3}]
        >>> filter_dicts(test_d, a=1, b=3, d=3)
        []
        >>> filter_dicts(test_d)
        [{'a': 1, 'b': 2}, {'a': 0, 'b': 3}, {'a': 1, 'd': 3}, {'a': 2, 'd': 3}]
        >>> filter_dicts([{}])
        [{}]
        >>> type(filter_dicts(test_d, as_list=False, a=1))
        <class 'filter'>
        >>> sum(1 for _ in filter_dicts(test_d, as_list=False, a=1))
        2
    """
    filtered_records = filter(
        lambda rec: all([rec.get(key) == value for key, value in filters.items()]), records)
    if as_list:
        filtered_records = list(filtered_records)
    return filtered_records


def filter_objects(objects: List[object], as_list=True, **filters) -> Union[List[object], Iterable[Dict]]:
    """
    Pass a list of objects and filters as kwargs to get all filtered records as list or filter obj

    Args:
        objects: list of objects to iterate and filter from
        as_list: Flag that shows if found records should be returned as list
        **filters: kwargs to be used for filtering as key:value pairs
    Returns:
        Filtered records as filter obj or list
    Examples:
        >>> class A:
        ...     def __init__(self, var1, var2):
        ...         self.var1, self.var2 = var1, var2
        ...     def __repr__(self):
        ...         return f'{self.var1}-{self.var2}'
        ...
        >>> filter_objects([A('test', 1), A('test2', 2)],  var1='test', var2=1)
        [test-1]
        >>> filter_objects([A('test', 1), A('test2', 2)], var1='test', var2=2)
        []
        >>> filter_objects([A('test', 1), A('test2', 2)])
        [test-1, test2-2]
        >>> filter_objects([{}])
        [{}]
        >>> type(filter_objects([A('test', 2), A('test2', 2)], as_list=False, var2=2))
        <class 'filter'>
        >>> sum(1 for _ in filter_objects([A('test', 2), A('test2', 2)], as_list=False, var2=2))
        2
    """
    filtered_objects = filter(
        lambda rec: all([getattr(rec, key) == value for key, value in filters.items()]), objects)
    if as_list:
        filtered_objects = list(filtered_objects)
    return filtered_objects


def find_object(objects: List[object], **filters) -> Optional[object]:
    """
    Pass a list of objects and filters as kwargs to get first occurence record. If no filters
    passed return first object in the list

    Args:
        objects: list of objects to iterate and filter from
        **filters: kwargs to be used for filtering as key:value pairs
    Returns:
        Found record obj or None
    Examples:
        >>> class A:
        ...     def __init__(self, var1, var2, var3=False):
        ...         self.var1, self.var2, self.var3 = var1, var2, var3
        ...     def __repr__(self):
        ...         return f'{self.var1}-{self.var2}'
        ...
        >>> find_object([A('test', 1), A('test2', 2)],  var1='test', var2=1)
        test-1
        >>> find_object([A('test', 1), A('test2', 2)], var1='test', var2=2) is None
        True
        >>> find_object([{}])
        {}
        """
    for rec in objects:
        if all([getattr(rec, key) == value for key, value in filters.items()]):
            return rec


def exclude_keys(dictionary: Dict, keys: List[Hashable]) -> Dict:
    """
    Create a new dictionary, excluding the keys given.

    Args:
        dictionary: Source dict
        keys: list of keys which we don't want to include in the new dict
    Examples:
        >>> exclude_keys({'a': 1, 'b': 2}, keys=['a'])
        {'b': 2}
    """
    return {k: v for k, v in dictionary.items() if k not in keys}


def keep_keys(dictionary: Dict, keys: List[Hashable]) -> Dict:
    """
    Create a new dictionary, keeping only the given keys.

    Args:
        dictionary: Source dict
        keys: list of dict keys which we want to include in the new dict
    Examples:
        >>> keep_keys({'a': 1, 'b': 2}, keys=['a'])
        {'a': 1}
    """
    return {k: v for k, v in dictionary.items() if k in keys}


def exclude_none_values(dictionary: Dict) -> Dict:
    """
    Create a new dictionary, removing the keys whose value is None.

    Args:
        dictionary: Source dict
    Examples:
        >>> exclude_none_values({'a': None, 'b': 1})
        {'b': 1}
    """
    return {k: v for k, v in dictionary.items() if v is not None}


def get_values(dictionary: Dict, keys: List[Hashable], dtypes: Dict = None) -> Dict:
    """
    Get values from dictionary whose keys are in the keys list

    Args:
        dictionary: Dictionary with the values
        keys: List of keys to extract
        dtypes: A mapping of fields to types, used to convert fields
    Returns:
        New dict with the key:value pairs
    Examples:
        >>> get_values({'a': '1', 'b': 2, 'c': '3'}, keys=['a', 'c'], dtypes={'a': int})
        {'a': 1, 'c': '3'}
    """
    data = {}
    for key in keys:
        value = dictionary.get(key)
        # Apply type conversion
        if dtypes and value is not None and dtypes.get(key):
            value = dtypes[key](value)
        data[key] = value
    return data


def get_differences(
        old_data: DictOrObject,
        new_data: Dict,
        skip_keys: List[Hashable] = None,
        number_precision: int = 6
) -> Dict:
    """
    Get a dictionary with the values that have changed between two versions of data.

    Args:
        old_data: Object or dictionary containing the old version of the data
        new_data: Dictionary containing the new version of the data
        skip_keys: Optional list of keys to skip during comparison
        number_precision: Precision used for number comparisons
    Returns:
        Dict containing the keys that have changed with the new respective values
    """
    differences = {}
    if not skip_keys:
        skip_keys = []

    old_data_is_dict = isinstance(old_data, dict)

    for key, new_value in new_data.items():
        if key in skip_keys:
            continue

        if old_data_is_dict:
            old_value = old_data[key]
        else:
            old_value = getattr(old_data, key)

        # Process Decimal
        # Process datetime - Add UTC timezone when missing
        if isinstance(old_value, datetime) and not old_value.tzinfo:
            old_value = old_value.replace(tzinfo=timezone.utc)
        if isinstance(new_value, datetime) and not new_value.tzinfo:
            new_value = new_value.replace(tzinfo=timezone.utc)

        if isinstance(new_value, (float, Decimal)):
            # Compare numbers taking precision into consideration
            if (old_value is None
                    or not are_equal(old_value, new_value, precision=number_precision)):  # type: ignore
                differences[key] = new_value

        # Compare values
        else:
            if old_value != new_value:
                differences[key] = new_value

    return differences


def have_equal_values(dict1: Dict, dict2: Dict, attributes: List = None) -> bool:
    """
    Compare if the given attributes of two dictionaries are equal.

    Args:
        dict1: dict to be compared
        dict2: dict to be compared against
        attributes: list of keys that are going to compare, default None, check all keys in dict1
    Returns:
        True if dict1 attributes are equal with dict2, False otherwise
    Examples:
        >>> have_equal_values({'a': 1, 'b': 3}, {'a': 1, 'b': 2}, attributes=['a'])
        True
        >>> have_equal_values({'a': 1, 'b': 3}, {'a': 1, 'b': 2})
        False
    """

    if not attributes:
        attributes = dict1.keys()

    for attribute in attributes:
        if dict1.get(attribute) != dict2.get(attribute):
            return False

    return True


def get_nested(dictionary: Dict, *attrs: Hashable) -> Optional[Any]:
    """
    Access a nested value in a dict by passing the keys of all the levels.

    Args:
        dictionary: Dict object we want to access
        *attrs: Keys we want to access
    Returns:
        The value we want to get or None if it can't be found
    Raises:
        AttributeError: When the original object or a nested one is not a dict
    Examples:
        >>> get_nested({'a': {'b': {'c': 'Value'}}}, 'a', 'b', 'c')
        'Value'
        >>> get_nested({}, 'a') is None
        True
        >>> get_nested({'a': 'b'}, 'c') is None
        True
        >>> get_nested({'a': {'b': {'c': 'Value'}}}, 'a', 'd', 'c') is None
        True
        >>> get_nested(1, 'a') is None  # type: ignore
        True
    """
    if not dictionary or not attrs:
        return None

    current_value = dictionary
    for attr in attrs:
        try:
            current_value = current_value.get(attr)
        except AttributeError:
            return None
    return current_value


def lists_intersection(list1: List, list2: List) -> List:
    """
    Find the common elements between 2 lists.

    Args:
        list1: list to look for intersection
        list2: list to be looked against
    Returns:
        new list with the intersection of the 2 lists
    Examples:
        >>> lists_intersection([1, 2, 3], [2, 3, 4])
        [2, 3]
        >>> lists_intersection([1, 2, 3], [4, 5, 6])
        []
        >>> lists_intersection([1, 'test', 3], [4, 5, 'test'])
        ['test']
    """
    return list(set(list1).intersection(list2))


def chunks(lst: List, chunk: int) -> Generator[List, None, None]:
    """
    Yield successive n-sized chunks from lst.

    Args:
        lst: list to yield chunks from
        chunk: size of the chunk
    Returns:
        list of elements from generator
    Examples:
        >>> a = chunks([1, 2, 3, 4, 5, 6], chunk=2)
        >>> print(type(a))
        <class 'generator'>
        >>> sum(1 for _ in a)
        3
        >>> next(chunks([1, 2, 3, 4, 5, 6], chunk=2))
        [1, 2]
    """
    for i in range(0, len(lst), chunk):
        yield lst[i: i + chunk]


def ints(values: List[str]) -> List[int]:
    """
    Converts elements of a list to ints

    Args:
        values: list of elements to be converted
    Returns:
        list with ints representations of elements
    Examples:
        >>> ints(['1', '2', '3'])
        [1, 2, 3]
    """
    return [int(value) for value in values]


def any_not_none(iterable: Iterable) -> bool:
    """
    Verify if any of the elements of the iterable is not None.
    The default behaviour of the builtin any function checks the value
    with `if element` and causes the values like zero (0) to be
    treated as Falsy values. This function aims to change this behaviour
    to return False only when all the values are None.

    Args:
        iterable: Iterable in which we will look for not None values
    Returns:
        bool value indicating if at least one element is not None
    Examples:
        >>> any_not_none([None, None, 13])
        True
        >>> any_not_none([None, None, 0])
        True
        >>> any_not_none([None])
        False
        >>> any_not_none([])
        False
    """
    for element in iterable:
        if element is not None:
            return True

    return False


class Finder:
    """
    Finder class to be used for imitating an ORM with dict and objects
    Examples:
        >>> records = [{'id': 1, 'foo': 'bar'}, {'id': 2, 'a': 'b'}]
        >>> Finder(records).find(id=1)
        {'id': 1, 'foo': 'bar'}
        >>> Finder(records).filter(foo='bar')
        [{'id': 1, 'foo': 'bar'}]
    """
    ops = {
        "gt": operator.gt,
        "gte": operator.ge,
        "lt": operator.lt,
        "lte": operator.le,
        "eq": operator.eq,
        "ne": operator.ne,
        "in": lambda item, iterable: item in iterable,
    }
    available_ops = ops.keys()

    def __init__(self, records: List[DictOrObject]):
        self._records = records

    @classmethod
    def _compare(cls, value1: Any, op: str, value2: Any, ignore_types=False) -> bool:
        if ignore_types:
            value1 = str(value1)
            value2 = str(value2)
        return cls.ops[op](value1, value2)

    @classmethod
    def _verify(cls, record: DictOrObject, ignore_types=False, **checks) -> bool:
        """
        Verify that the record fulfills the given checks.

        Args:
            record: The record to be verified: dict or object
            ignore_types: If True, compare the values as strings
            **checks: Dictionary with checks in the form: {"attr__gt": 5}

        Returns:
            True if the record passes *all* the checks, False otherwise
        """
        for key, value in checks.items():
            if "__" in key:
                elements = key.split("__")
                # If operator has been declared
                if elements[-1] in cls.available_ops:
                    attr = ".".join(elements[:-1])
                    op = elements[-1]
                # eq operator by default
                else:
                    attr = ".".join(elements)
                    op = "eq"
            else:
                attr = key
                op = "eq"
            if not cls._compare(cls._get_value(record, attr), op, value, ignore_types):
                return False
        return True

    @staticmethod
    def _get_value(record: DictOrObject, attr: str) -> Optional[Any]:
        """
        Get the value of the attribute for the given record. Used to process dicts and objects uniformly.

        Args:
            record: The record from which the value will be retrieved. Can be a dict or an object
            attr: The attribute to retrieve value. For objects, supports syntax like 'asset__debtor__country_id'.
        Returns:
            Value found or None
        """
        if isinstance(record, dict):
            return record.get(attr)
        else:
            try:
                while "." in attr:
                    current_attr, attr = attr.split(".", 1)
                    record = getattr(record, current_attr)
                return getattr(record, attr)
            except AttributeError:
                return None

    def filter(self, as_list=True, ignore_types=False, **filters):
        filtered_records = filter(lambda rec: self._verify(rec, ignore_types, **filters), self._records)
        if as_list:
            filtered_records = list(filtered_records)
        return filtered_records

    def find(self, **filters):
        for record in self._records:
            if self._verify(record, **filters):
                return record



import math
from decimal import Decimal
from numbers import Number
from typing import Optional, Any


def dec_multiply(*args) -> Optional[Decimal]:
    """
    Multiplication of numbers passed as *args.

    Args:
        *args: numbers we want to multiply
    Returns:
        The result of the multiplication as a decimal number
    Examples:
        >>> dec_multiply(3, 3.5, 4, 2.34)
        Decimal('98.280')
        >>> dec_multiply() is None
        True
    """
    if not args:
        return
    total = Decimal(str(args[0]))
    for element in args[1:]:
        total *= Decimal(str(element))
    return total


def dec_sum(*args: Optional[Number]) -> Decimal:
    """
    Sum of numbers passed as *args. Skips None values

    Args:
        *args: numbers we want to sum
    Returns:
        The result of the sum as a decimal number
    Examples:
        >>> dec_sum(20.24, 2.2)
        Decimal('22.44')
        >>> dec_sum(20.24, None, 0, 10)
        Decimal('30.24')
    """
    return sum([Decimal(str(x)) if x is not None else 0 for x in args])


def dec_ratio(value1: Number, value2: Number) -> Optional[Decimal]:
    """
    Decimal ratio against 2 numbers. If value2 is 0 return None

    Args:
        value1: First number, numerator
        value2: Second number, denominator
    Returns:
        The result of ratio as a decimal number
    Examples:
        >>> dec_ratio(20.24, 2.2)
        Decimal('9.2')
        >>> dec_ratio(20.24, 0) is None
        True
    """
    return Decimal(str(value1)) / Decimal(str(value2)) if value2 else None


def dec_subtraction(value1: Optional[Number], value2: Optional[Number]) -> Decimal:
    """
    Decimal subtraction against 2 numbers

    Args:
        value1: First number
        value2: Second number as subtractor
    Returns:
        The result of subtraction as a decimal number
    Examples:
        >>> dec_subtraction(14.3, 3.567)
        Decimal('10.733')
        >>> dec_subtraction(None, None)
        Decimal('0')
    """
    return Decimal(str(value1) if value1 else 0) - Decimal(str(value2) if value2 else 0)


def parse_number(v: str) -> float:
    """
    Parse a given text which may contain commas and dots to float.

    Args:
        v:  str, value to be parsed to float
    Returns:
        v:  float, parsed value
    Examples:
        >>> parse_number('0,00')
        0.0
        >>> parse_number('23.789,90')
        23789.9
        >>> parse_number('23789.80')
        23789.8
        >>> parse_number('23,789.80')
        23789.8
        >>> parse_number('234,234,342.80')
        234234342.8
    """
    if ',' in v:
        if '.' in v:
            if v.find('.') < v.find(','):
                v = float(v.replace('.', '').replace(',', '.'))
            else:
                v = float(v.replace(',', ''))
        else:
            v = float(v.replace(',', '.'))
    else:
        v = float(v)

    return v


def safe_min(*values) -> Optional[Any]:
    """
    Find the min value in a list. Ignore None values.

    Args:
        *values: all values to be compared
    Returns:
         min value in the list or None
    Examples:
        >>> safe_min(None, 5, 2, 1, None, 5)
        1
        >>> safe_min(None, None) is None
        True
    """
    min_value = None
    for value in values:
        if value is None:
            continue
        if min_value is None:
            min_value = value
        elif value < min_value:
            min_value = value
    return min_value


def safe_max(*values) -> Optional[Any]:
    """
    Find the max value in a list. Ignore None values.

    Args:
        *values: all values to be compared
    Returns:
         max value in the list or None
    Examples:
        >>> safe_max(None, 5, 3, 7)
        7
        >>> safe_max(None, None) is None
        True
    """
    max_value = None
    for value in values:
        if value is None:
            continue
        if max_value is None:
            max_value = value
        elif value > max_value:
            max_value = value
    return max_value


def are_equal(number1: Number, number2: Number, precision=6) -> bool:
    """
    Compare two decimal numbers with a given precision

    Args:
        number1: number to be compared
        number2: number which will be used as the comparator
        precision: represent what precision to use when comparing decimal numbers
    Returns:
        True if both are equal or False otherwise
    Examples:
        >>> are_equal(13.44, 13.4, precision=1)
        True
        >>> are_equal(13.44, 13.4)
        False
    """
    return abs(dec_subtraction(number1, number2)) < Decimal(str(1 / (10 ** precision)))


def evaluate(expression: str):
    """
    Evaluate a math expression.

    Args:
        expression: string representing the math expression to be eval()
    Returns:
        result of the math expression from eval()
    Raises:
        NameError: If expression passed is not from math library
    Examples:
        >>> evaluate("5 + 2")
        7
        >>> evaluate("sqrt(9)")
        3.0
        >>> evaluate("hello - hello")
        Traceback (most recent call last):
        ...
        NameError: The use of 'hello' is not allowed
    """

    # Compile the expression
    code = compile(expression, "<string>", "eval")
    allowed_names = {k: v for k, v in math.__dict__.items() if not k.startswith("__")}

    # Validate allowed names
    for name in code.co_names:
        if name not in allowed_names:
            raise NameError(f"The use of '{name}' is not allowed")
    return eval(code, {"__builtins__": {}}, allowed_names)



from urllib.parse import urlencode

from rest_framework.response import Response
from rest_framework.status import HTTP_400_BAD_REQUEST, HTTP_401_UNAUTHORIZED, HTTP_403_FORBIDDEN, \
    HTTP_502_BAD_GATEWAY


def bad_request(message="Bad request."):  # pragma no cover
    return Response(data={"message": message}, status=HTTP_400_BAD_REQUEST)


def unauthorized(message="Unauthorized."):  # pragma no cover
    return Response(data={"message": message}, status=HTTP_401_UNAUTHORIZED)


def forbidden(message="Forbidden."):  # pragma no cover
    return Response(data={"message": message}, status=HTTP_403_FORBIDDEN)


def bad_gateway(message="Bad Gateway."):  # pragma no cover
    return Response(data={"message": message}, status=HTTP_502_BAD_GATEWAY)


def add_params_to_url(url: str, **kwargs) -> str:
    """
    Join the url and params ready for making requests

    Args:
        url: string representing the url to be called
        **kwargs: all the params to join the url
    Returns:
        Final url ready for request
    Examples:
        >>> add_params_to_url('https://test.com', p1='20', p2=30, p3=None, p4='test', p5='<html/>')
        'https://test.com?p1=20&p2=30&p3=None&p4=test&p5=%3Chtml%2F%3E'
    """
    return url if not kwargs else f"{url}?{urlencode(kwargs)}"



def re_raise_exception(exception: Exception, *new_args):
    """
    Get an Exception obj and reraise it with extra new passed arguments

    Args:
        exception: Exception obj to be raised
        *new_args: Args to be passed to the exception obj
    Raises:
        The exception obj with the new args attached
    Examples:
        >>> re_raise_exception(Exception('Old exception:'), 'new arg', 'another one')
        Traceback (most recent call last):
        ...
        Exception: ('Old exception:', 'new arg', 'another one')
        >>> re_raise_exception(Exception('Old exception:'))
        Traceback (most recent call last):
        ...
        Exception: Old exception:
    """
    exception.args += tuple(new_args)
    raise exception
